#!/usr/bin/env ruby

require 'z80'

module Opcode
	RET  = 0xC9
	NOP  = 0x00
	CALL = 0xCD
end

quit = cursor_x = tab = memory = nil
cpu = Z80.new

cpu.fetch_opcode = cpu.fetch = cpu.read do |context, address|
	memory[address]
end

cpu.write do |context, address, value|
	memory[address] = value if address > 0x3FFF
end

cpu.in do |context, port|
	port.odd? ? 255 : 191
end

cpu.hook do |context, address|
	case address
	when 0x0010 # PRINT
		if tab == 0
			case (c = cpu.a)
			when 0x0D # CR
				putc "\n"
				cursor_x = 0
			when 0x17 # TAB
				tab = 2
			when 0x7F # ©
				print "©"
				cursor_x += 1
			else
				if c >= 32 && c < 127
					putc c
					cursor_x += 1
				end
			end
		elsif (tab -= 1) != 0
			c = 0x1F & cpu.a
			x = 0x1F & cursor_x
			if c < x
				putc "\n"
				cursor_x = 0
			else
				cursor_x += (c -= x)
			end
			print ' ' * c
		end
		Opcode::RET
	when 0x7003 # Exit
		cpu.terminate
		quit = true
		Opcode::NOP
	else
		Opcode::NOP
	end
end

$stdout.sync = true if $stdout.tty?

ARGV.each do |file_path|
	program = file_path == '-' ? $stdin.read : File.read(file_path)
	puts "#{file_path}:"
	quit     = false
	cursor_x = tab = 0
	memory   = Array.new(65536, 0)
	memory[0x8000, program.size - 91] = program.bytes[91..-1]
	memory[0x0010] = Z80::HOOK    # THE 'PRINT A CHARACTER' RESTART
	memory[0x0D6B] = Opcode::RET  # THE 'CLS' COMMAND ROUTINE
	memory[0x1601] = Opcode::RET  # THE 'CHAN_OPEN' SUBROUTINE
	memory[0x7000] = Opcode::CALL # -.
	memory[0x7001] = 0x00         #  |- call 8000h
	memory[0x7002] = 0x80         # -'
	memory[0x7003] = Z80::HOOK
	cpu.power true
	cpu.im = 1
	cpu.i  = 0x3F
	cpu.pc = 0x7000
	cpu.run(Z80::MAXIMUM_CYCLES) until quit
	break if file_path == '-'
end
